package main

import (
	"log"
	"os"
	"path/filepath"
	"strings"
	"syscall"

	"kernel.org/pub/linux/libs/security/libcap/cap"
)

func setFromCaps(caps ...string) (*cap.Set, error) {
	var err error
	newSet := cap.NewSet()
	value_list := make([]cap.Value, 0, len(caps))
	for _, cap_name := range caps {
		value, err := cap.FromName(strings.ToLower(strings.TrimSpace(cap_name)))
		if err != nil {
			return nil, err
		}
		value_list = append(value_list, value)
	}

	err = newSet.SetFlag(cap.Effective, true, value_list...)
	if err != nil {
		return nil, err
	}
	err = newSet.Fill(cap.Permitted, cap.Effective)
	if err != nil {
		return nil, err
	}
	return newSet, err
}

func ChownRecursively(root string, owner int, group int) {
	err := filepath.Walk(root,
		func(path string, info os.FileInfo, err error) error {
			if err != nil {
				return err
			}
			if int(info.Sys().(*syscall.Stat_t).Uid) != owner || int(info.Sys().(*syscall.Stat_t).Gid) != group {
				err = os.Chown(path, owner, group)
				if err == nil {
					log.Printf("File ownership of %s changed.\n", path)
				}
			}
			return nil
		})
	if err != nil {
		log.Printf("Chown failed: %v", err)
	}
}

// mustDropPrivileges executes the program in a child process, dropping root
// privileges, but retaining the CAP_NET_BIND_SERVICE capability to binding to TCP/UDP sockets below 1024.
func mustDropPrivileges(uid int, gid int, caps ...string) {
	// Read and display the capabilities of the running process
	c := cap.GetProc()
	if os.Getenv("GITEA_PRIVILEGES_DROPPED") == "1" {
		return
	}
	capSet, err := setFromCaps(caps...)
	if err != nil {
		log.Fatalf("Unable to parse capabilities: %s", err)
	}
	diff, _ := capSet.Cf(c)
	if gid == os.Getgid() && uid == os.Getuid() && diff == 0 {
		return
	}
	os.Setenv("GITEA_PRIVILEGES_DROPPED", "1")
	ChownRecursively(os.Getenv("HOME"), uid, gid)

	if gid != 0 {
		err = cap.SetGroups(gid)
		if err != nil {
			log.Fatalf("Unable to change gid to %d:", gid, err)
		}
	}
	if gid != 0 {
		err = cap.SetUID(uid)
		if err != nil {
			log.Fatalf("Unable to change uid to %d:", uid, err)
		}
	}

	err = capSet.SetProc()
	if err != nil {
		log.Fatalf("Unable to drop privileges: %s", err)
	}

}
